'''
Created on Apr. 28, 2019

@author: jldupont
'''
import unittest

from jldjson.tools import unpack

class Test(unittest.TestCase):

    def testUnpack1(self):
        d={"a": 666, "b": {"b1": 777, "b2": 888}}

        r=unpack(d, ["b"])
        
        assert r['a'] == 666
        assert r['b1'] == 777
        assert r['b2'] == 888
        
        assert r.get("b", None) == None

if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    unittest.main()


'''
Created on Apr. 28, 2019

@author: jldupont

Performs various operations on NL JSON stdin stream

'''
__version__ = "0.1"

import sys
import click
import json

import jldjson.tools as tools

@click.command()
@click.option('--version', '-v', help="Show version", is_flag=True)
@click.option('--unpack', '-u', help="Unpack a dictionary", multiple=True)
@click.option('--ignore', '-i', help="Ignore error", default=False, is_flag=True)
@click.option('--stderr', help="Errors will be printed to sys.stderr", is_flag=True)
@click.option('--keep', '-k', multiple=True, help="Key to keep in input object")
def command(version, unpack, ignore, stderr, keep):
    """
    Various operators to JSON NL stdin stream
    """
    if version:
        print("Version: ", __version__)
        sys.exit(0)
    
    for line in sys.stdin:
        
        try:
            jobj = tools.loader(line)
            to_unpack = tools.keep(jobj, keep)
            out = tools.unpack(to_unpack, unpack)
                        
        except tools.Skip:
            continue
                        
        except Exception as e:
            out = None
            
            if stderr:
                print("Error: ", str(e), file=sys.stderr)
            
            if ignore == False:
                raise e
            
        if out is None:
            continue

        print(json.dumps(out))
    

if __name__ == '__main__':
    command()


'''
Created on Apr. 24, 2019

@author: jldupont

state(a,b,c) = action
'''

class StateExists(Exception): pass
class NoDefaultStateExists(Exception): pass

class States:
    """
    Collection of [state, action] pairs
    
    The hierarchical map is constructed using the following:
    
      { l1_state1: { l2_state1: { l3_state1: action_state1 } }
       ,l1_state2: { l2_state2: { l3_state2: action_state2 } }
       ,l1_state3: { l2_state3: action_state3 }
       ... 
      }
      
    Whereas 'action' is a callable object and 'state' is a string.
    """
    
    def __init__(self):
        self._map = {}
    
    def add(self, action, *components):
        """
        Add a [state, action] pair
        
        Each positional parameter constitute
        a level in the hierarchy of the map
        """
        assert len(components) > 0, "Must have at least of state component"
        assert callable(action), "Parameter 'action' must be a callable"
        
        def _add(_map, action, cp):
            
            # Terminate the recursion
            if len(cp) == 0:
                return
            
            # Are we at the last step, ready to assign the action?
            if len(cp) == 1:
                #
                # Is there already something at this state?
                #
                try:
                    _ = _map[cp[0]]
                except:
                    #
                    # State does not exists so use, perfect.
                    #
                    _map[cp[0]] = action
                else:
                    raise StateExists("State is already defined at: "+str(components))
                
                return
            
            #
            # We are in the "middle" of the process.
            # 
            #
            down_map = _map.get(cp[0], {})
            
            if callable(down_map):
                raise StateExists("Action already present at this state: "+str(components))
            
            #
            # This is just really necessary to initialize the chain
            #
            _map[cp[0]] = down_map
            
            # Recurse...
            #
            _add(down_map, action, cp[1:])
        
        _add(self._map, action, components)
        
    def get(self, *components):
        """
        Retrieves a mapping [state, action]
        """
        assert len(components) > 0, "Must have at least 1 state component"
        
        def _drill(_map, cp):
            comp = cp[0]
            if len(cp) == 1:
                maybe_action = _map.get(comp, None)
                if maybe_action is None:
                    return None
                
                if callable(maybe_action):
                    return maybe_action
                
                #
                # It's not a callable
                #  so it can't be a match at this point...
                #
                return None
            
            maybe_next_level = _map.get(comp, None)
            if maybe_next_level is None:
                return None
            
            return _drill(maybe_next_level, cp[1:])
        
        return _drill(self._map, components)
    
    def step(self, context, *components):
        """
        Performs a step
        """
        action = self.get(*components)
        
        if action is None:
            catch_all_entry = list(components[:-1])
            catch_all_entry.append('*')
            # Is there a catch-all in place ?
            default_action = self.get(*catch_all_entry)
            if default_action is None:
                raise NoDefaultStateExists(str(catch_all_entry))
            
            action = default_action
        
        return action({"context": context, "state": components})

    


'''
Created on Apr. 28, 2019

@author: jldupont
'''
import json


class ExpectingDict(Exception):
    """
    The value in the (key:value) pair
    is not of type dict
    """

class ExpectingJSONObject(Exception):
    """
    Expecting to decode a JSON object
    from the provided string
    """

class ControlFlow(Exception):
    """
    Base class for flow control exceptions
    """

class Skip(ControlFlow):
    """
    Skip the current iteration
    """


def loader(line):
    
    line = " ".join(line.split())
    
    if len(line) == 0:
        raise Skip()
    
    try:
        jobj = json.loads(line)
        
    except:
        raise ExpectingJSONObject("Got: "+str(line))
    
    return jobj


def keep(input_obj, which_key_to_keep):
    out = {}
    for key in which_key_to_keep:
        if key in input_obj:
            out[key] = input_obj[key]
    return out
    

def unpack(input_obj, which_key_to_unpack):
    
    out = {}
    
    for key, value in input_obj.items():
        
        if key in which_key_to_unpack:
            if not isinstance(value, dict):
                raise ExpectingDict("key: "+key)
            
            out.update(input_obj[key])
            
        else:
            out[key] = value
        
    return out



'''
Converts geojson FeatureCollection to Newline delimited geojson geometry features

Created on Apr. 23, 2019

@author: jldupont

Example geojson FeatureCollection object

{
"type": "FeatureCollection",
"name": "ON_LDU",
"crs": { "type": "name", "properties": { "name": "urn:ogc:def:crs:OGC:1.3:CRS84" } },
"features": [
{ "type": "Feature", "properties": { "PCA_ID": 180000059, "POSTALCODE": "P0R1L0", "PROV": "ON", "MAF_ID": 3500025
23, "PREC_CODE": 2, "PCA_COUNT": 516, "DOM_PCA": 0, "MULTI_PC": 0, "DEL_M_ID": "LB0001", "LONGITUDE": -83.4425680
69900005, "LATITUDE": 46.289068144200002 }, "geometry": { "type": "MultiPolygon", "coordinates": [ [ [ [ -83.4323
79699999956, 46.29642750000005 ], [ -83.432314099999985, 46.295825900000068 ], [ -83.4323528, 46.29552810000007 ]
, [ -83.4323506, 46.294731200000058 ], [ -83.432348399999967, 46.293962100000044 ], [ -83.432283299999938, 46.293
666500000029 ], [ -83.432166, 46.293363 ], ... ] ] ] } },

Algorithm:
* features start_array

* features.item start_map 
   ==> features.item.properties

* features.item end_map

'''
import sys
import logging
import click
import ijson

from jldgeo.fsm import GeojsonFsm, EndOfFile

@click.command()
@click.option(
    '--inputfile'
    ,type=click.File('r')
    ,help="""File of geojson format containing a FeatureCollection object"""
    ,required=True
)
@click.option(
    '--count'
    ,default=None
    ,help="""The maximum number of objects to emit on stdout"""
)
def geojson_converter(inputfile, count):
    """
    The main entry point
    """
    if inputfile is None:
        raise Exception("Missing ")
    
    max_count = count
    current_count = 0
    parser = ijson.parse(inputfile)
    
    f = GeojsonFsm()
    
    for prefix, event, value in parser:
        try:
            f.submitEvent(prefix, event, value)
            
        except EndOfFile:
            sys.exit(0) # no error
            break
        
        except Exception as e:
            logging.error(e)
            sys.exit(1)
        
        if count != None:  
            current_count += 1
            if current_count == max_count:
                break
    

if __name__ == '__main__':
    geojson_converter()